/*********************************************************************/
/* neblo.h                                                           */
/* types, constants, and prototypes for neblo                        */
/* see COPYRIGHT in parent directory                                 */
/*********************************************************************/

#ifndef NEBLO_API_H
#define NEBLO_API_H


#include <sys/types.h>       /* standard system types                */

/* max size for a neblo message payload                              */
/* it must hold HEADERSIZE_F + NEBLO_MSGSIZE <= KEYBUFLEN            */
#define NEBLO_MSGSIZE  65536

/* max redundancy for messages                                       */
#define NEBLO_MAXCP        5


/* two different waiting semantics for single pending REQuests       */
#define NEBLO_WAITALL   0
#define NEBLO_WAITANY   1


typedef u_int64_t neblo_addr;
typedef u_int16_t neblo_port;


/*  Handlers for messages  */

typedef int ( *handler ) ( neblo_port, neblo_addr, void *, unsigned int * );


/*  Handlers for peer arrival/departures  */

typedef int ( *manager ) ( neblo_addr, neblo_addr );


/*
 *  Handles (not handler!) for request-response msgs,
 *  to wait/test for responses
 */

typedef u_int32_t neblo_rhandle;
#define R_NONE ((neblo_rhandle)(-1))


/*  Initialization routine  */

void neblo_init ( int argc, char **argv );


/*
 *  Attach a handler to a port, to handle UniDirectional msgs at a recipient
 *  or at a forwarder node.
 *  The originator is also a forwarder, if it is not also the recipient.
 */

void neblo_set_receiver_handler_UD ( neblo_port, handler );

void neblo_set_forwarder_handler_UD ( neblo_port, handler );


/*
 *  Attach a handler to a port, to handle REQuest-RESponse msgs at a recipient
 *  or at a forwarder.
 *  The originator is also a forwarder, if it is not also the recipient.
 *  At the recipient, the handler should compute an answer so "msg_len" might
 *  change.
 */

void neblo_set_receiver_handler_REQ ( neblo_port, handler );

void neblo_set_forwarder_handler_REQ ( neblo_port, handler );

void neblo_set_forwarder_handler_RES ( neblo_port, handler );


/*
 *  Set the migration and regeneration handlers,
 *  to manage node arrivals/departures.
 */

void neblo_set_refresh_handler ( manager );

void neblo_set_clean_handler ( manager );


/*  Connect to the overlay, after having set up the handlers.  */

int neblo_connect ( void );


/*  Send a UniDirectional msg.  No answer will be provided by the recipient  */

int neblo_UD_send
(
  unsigned int num_copies, neblo_port port, neblo_addr addr, void *msg,
  unsigned int len
);


/*  A variant of neblo_UD_send(), for exclusive use in handlers  */

int neblo_UD_send_h
( neblo_port port, neblo_addr addr, void *msg, unsigned int len );


/*
 *  Send a REQuest-RESponse msg.  An answer will be provided by the recipient.
 *  "msg_len" is the REQuest len, "res_maxlen" is the max len allowed for the
 *  expected RESponse.
 *  Semantics is non-blocking with a split-phase architecture, so wait/test
 *  routines must be invoked to synchronize with the response; "rhandle"
 *  binds the two phases with each other.
 */

int neblo_REQ_send
(
  unsigned int num_copies, neblo_port port, neblo_addr addr,
  void *msg, unsigned int msg_len, void *res[], unsigned int res_maxlen,
  neblo_rhandle *rhandle
);

/*  A variant of neblo_REQ_send(), for exclusive use in handlers  */

int neblo_REQ_send_h
(
  neblo_port port, neblo_addr addr,
  void *msg, unsigned int msg_len, void *res[], unsigned int res_maxlen,
  neblo_rhandle *rhandle
);

int neblo_RES_wait ( neblo_rhandle, int );

int neblo_RES_test ( neblo_rhandle, int );


/*  Miscellaneous  */

unsigned int neblo_in_range ( neblo_addr,  neblo_addr,  neblo_addr );

int neblo_my_id ( neblo_addr * );

void neblo_lock ( void );

int neblo_trylock ( void );

void neblo_unlock ( void );


/*  Other, "unofficial" utilities  */

/*
 *  Fold a "long" word (src) over a shorter one (dest) by xor.
 *  Error if src is actually shorter than dest.
 */
static inline int
xor_fold
( int8_t *dest, unsigned int dest_size, int8_t *src, unsigned int src_size )
{

  unsigned int  ds;
  int8_t       *d;


  if (src_size < dest_size) {
//    fprintf(stderr,"xor_fold():  error: source shorter than dest\n");
    return -1;
  }
  memcpy(dest,src,dest_size);
  if (src_size == dest_size) {
    /*  No need to fold  */
    return 0;
  }
  /*  Fold over in chunks, each dest_size long.  First chunk already there.  */
  src += dest_size;
  src_size -= dest_size;
  d = dest;
  ds = dest_size;
  for (; src_size; src_size--, src++) {
    *d ^= *src;
    if (!--ds) {
      d = dest;
      ds = dest_size;
    } else
      d++;
  }
  return 1;

}


#endif
